#region LIOGO - GPL Copyright (c) 2005-2007 Lionel Lask
//
// This file is part of LIOGO.
//
// LIOGO is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// LIOGO is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Liogo; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//================================================================================
#endregion

using System;
using System.ComponentModel;
using System.Globalization;
using MyDLR;

namespace LIOGO.Core
{
	/// <summary>
	/// Common interface for all Logo Values (Word, Number and List).
	/// </summary>
    [TypeConverter(typeof(LogoConverter))]   // Added for MyDLR project
    public interface ILogoValue : ICallable    // ICallable added for MyDLR project
	{
		#region Test Operations
		/// <summary>
		/// Test if the value is a Word
		/// </summary>
		/// <returns>true if the value is a Word</returns>
		bool IsWord();

		/// <summary>
		/// Test if the value is a Number
		/// </summary>
		/// <returns>true if the value is a Number</returns>
		bool IsNumber();

		/// <summary>
		/// Test if the value is a List
		/// </summary>
		/// <returns>true if the value is a List</returns>
		bool IsList();

        /// <summary>
        /// Test if the value is an Array
        /// </summary>
        /// <returns>true if the value is an Array</returns>
        bool IsArray();

		/// <summary>
		/// Test if the value is empty
		/// </summary>
		/// <returns>true if the value is empty</returns>
		bool IsEmpty();

		/// <summary>
		/// Test if the value is true
		/// </summary>
		/// <returns>true if the value is equal to true</returns>
		bool IsTrue();

		/// <summary>
		/// Test if the value is false
		/// </summary>
		/// <returns>true if the value is equal to false</returns>
		bool IsFalse();
		#endregion

		#region Composite Operations
		/// <summary>
		/// Get item N in the value.
		/// </summary>
		/// <param name="index">Index of item (start at 1)</param>
		/// <returns>Value at index</returns>
		ILogoValue Item(ILogoValue index);

		/// <summary>
		/// Get the first item.
		/// </summary>
		/// <returns>First item</returns>
		ILogoValue First();

		/// <summary>
		/// Get the last item.
		/// </summary>
		/// <returns>Last item</returns>
		ILogoValue Last();

		/// <summary>
		/// Get everything except the first item.
		/// </summary>
		/// <returns>Everything except the first item</returns>
		ILogoValue ButFirst();

		/// <summary>
		/// Get everything except the last item.
		/// </summary>
		/// <returns>Everything except the last item</returns>
		ILogoValue ButLast();

		/// <summary>
		/// Put a new item at the first position.
		/// </summary>
		/// <param name="toPut">New item to put</param>
		/// <returns>A new value including the new item</returns>
		ILogoValue FPut(ILogoValue toPut);

		/// <summary>
		/// Put a new item at the last position.
		/// </summary>
		/// <param name="toPut">New item to put</param>
		/// <returns>A new value including the new item</returns>
		ILogoValue LPut(ILogoValue toPut);

		/// <summary>
		/// Get the length of the value.
		/// </summary>
		/// <returns>Length of the value</returns>
		int Length();
		#endregion

		#region Numeric Operations
		/// <summary>
		/// Add a value to the current value.
		/// </summary>
		/// <param name="toValue">Value to add</param>
		/// <returns>The result of sum</returns>
		ILogoValue Sum(ILogoValue toValue);

		/// <summary>
		/// Substract a value to the current value
		/// </summary>
		/// <param name="toValue">Value to substract</param>
		/// <returns>The result of difference</returns>
		ILogoValue Difference(ILogoValue toValue);

		/// <summary>
		/// Multiply a value by the current value
		/// </summary>
		/// <param name="toValue">Value to multiply</param>
		/// <returns>The result of product</returns>
		ILogoValue Product(ILogoValue toValue);

		/// <summary>
		/// Divide a value by the current value
		/// </summary>
		/// <param name="toValue">Value to divide</param>
		/// <returns>The result of quotient</returns>
		ILogoValue Quotient(ILogoValue toValue);

		/// <summary>
		/// Change sign of the current value.
		/// </summary>
		/// <returns>Negative of the value</returns>
		ILogoValue Minus();
		#endregion

		#region Comparison Operations
		/// <summary>
		/// Test if current value is equal to value.
		/// </summary>
		/// <param name="toValue">value to compare</param>
		/// <returns>true if values are equal</returns>
		bool IsEqual(ILogoValue toValue);

		/// <summary>
		/// Test if current value is not equal to value.
		/// </summary>
		/// <param name="toValue">value to compare</param>
		/// <returns>true if values are not equal</returns>
		bool NotEqual(ILogoValue toValue);

		/// <summary>
		/// Test if current value is greater than value.
		/// </summary>
		/// <param name="toValue">value to compare</param>
		/// <returns>true if current value is greater than value</returns>
		bool IsGreater(ILogoValue toValue);

		/// <summary>
		/// Test if current value is lower than value.
		/// </summary>
		/// <param name="toValue">value to compare</param>
		/// <returns>true if current value is lower than value</returns>
		bool IsLower(ILogoValue toValue);

		/// <summary>
		/// Test if current value is greater or equal to value.
		/// </summary>
		/// <param name="toValue">value to compare</param>
		/// <returns>true if current value is greater or equal to value</returns>
		bool IsGreaterOrEqual(ILogoValue toValue);

		/// <summary>
		/// Test if current value is lower or equal to value.
		/// </summary>
		/// <param name="toValue">value to compare</param>
		/// <returns>true if current value is lower or equal to value</returns>
		bool IsLowerOrEqual(ILogoValue toValue);
		#endregion

		#region Other Operations
		/// <summary>
		/// Do a copy of the value.
		/// </summary>
		/// <returns>a strict copy of the current value</returns>
		ILogoValue Copy();

		/// <summary>
		/// Convert the current value to a printable string.
		/// </summary>
		/// <returns>a string with the value</returns>
		string Print();

		/// <summary>
		/// Convert the current value to a showable string.
		/// </summary>
		/// <returns>a string with the value</returns>
		string Show();

		#endregion
	}
}
