#region LIOGO - GPL Copyright (c) 2005-2007 Lionel Lask
//
// This file is part of LIOGO.
//
// LIOGO is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// LIOGO is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Liogo; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//================================================================================
#endregion

using System;
using System.ComponentModel;
using System.Globalization;
using MyDLR;

namespace LIOGO.Core
{
	/// <summary>
	/// Class for a Logo Number.
	/// </summary>
    [TypeConverter(typeof(LogoConverter))]   // Added for MyDLR project
	public class LogoNumber : ILogoValue
	{
		/// <summary>
		/// Current value for the Number.
		/// </summary>
		private double currentValue;

        /// <summary>
        /// Neutral format for separators 
        /// </summary>
        private static NumberFormatInfo numberProvider;

        /// <summary>
        /// Initialize provider
        /// </summary>
        static LogoNumber()
        {
            numberProvider = new NumberFormatInfo();
            numberProvider.NumberDecimalSeparator = ".";
            numberProvider.NumberGroupSeparator = "";
        }
		
		#region Constructors
		/// <summary>
		/// Constructor with an int.
		/// </summary>
		/// <param name="initValue">Initial value</param>
		public LogoNumber(int initValue)
		{
			currentValue = (double)initValue;
		}

		/// <summary>
		/// Constructor with a double.
		/// </summary>
		/// <param name="initValue">Initial value</param>
		public LogoNumber(double initValue)
		{
			currentValue = initValue;
		}

		/// <summary>
		/// Constructor with a double.
		/// </summary>
		/// <param name="initValue">Initial value</param>
		/// <remarks>Throw an exception if conversion is impossible</remarks>
		public LogoNumber(ILogoValue initValue)
		{
			// Easy: it's a number
			if (initValue.IsNumber())
			{
				currentValue = (initValue as LogoNumber).currentValue;
				return;
			}

			// Try to convert to number
            if (!Double.TryParse(initValue.Print(), NumberStyles.Float, numberProvider, out currentValue))
                throw new Exception(initValue.ToString() + " must be a number");
		}

        /// <summary>
        /// Create a number from a string.
        /// </summary>
        /// <param name="value">init value</param>
        /// <returns>a LogoNumber or null if format don't match</returns>
        static public LogoNumber CreateFromString(string value)
        {
            double currentValue;
            if (!Double.TryParse(value, NumberStyles.Float, numberProvider, out currentValue))
                return null;
            return new LogoNumber(currentValue);
        }

        /// <summary>
        /// Create a number from a word.
        /// </summary>
        /// <param name="value">init value</param>
        /// <returns>a LogoNumber or null if format don't match</returns>
        static public LogoNumber CreateFromWord(LogoWord value)
        {
            return CreateFromString(value.Print());
        }
		#endregion

		#region Test Operations
		/// <summary>
		/// Test if the value is a Word
		/// </summary>
		/// <returns>true if the value is a Word</returns>
		public bool IsWord()
		{
			return false;
		}

		/// <summary>
		/// Test if the value is a Number
		/// </summary>
		/// <returns>true if the value is a Number</returns>
		public bool IsNumber()
		{
			return true;
		}

		/// <summary>
		/// Test if the value is a List
		/// </summary>
		/// <returns>true if the value is a List</returns>
		public bool IsList()
		{
			return false;
		}

        /// <summary>
        /// Test if the value is an Array
        /// </summary>
        /// <returns>true if the value is an Array</returns>
        public bool IsArray()
        {
            return false;
        }

		/// <summary>
		/// Test if the value is empty
		/// </summary>
		/// <returns>true if the value is empty</returns>
		public bool IsEmpty()
		{
			return false;
		}

		/// <summary>
		/// Test if the value is true
		/// </summary>
		/// <returns>true if the value is equal to true</returns>
		public bool IsTrue()
		{
			return false;
		}

		/// <summary>
		/// Test if the value is false
		/// </summary>
		/// <returns>true if the value is equal to false</returns>
		public bool IsFalse()
		{
			return false;
		}
		#endregion

		#region Composite Operations
		/// <summary>
		/// Get item N in the value.
		/// </summary>
		/// <param name="valueIndex">Index of item (start at 1)</param>
		/// <returns>Value at index</returns>
		public ILogoValue Item(ILogoValue valueIndex)
		{
			// Index must be a number or must be convertible to number
			LogoNumber index;
			if ( !valueIndex.IsNumber() )
			{
				if ( valueIndex.IsList() || valueIndex.IsArray() )
                    throw new Exception(valueIndex.ToString() + " must be a number");
				index = new LogoNumber(valueIndex);
			}
			else
				index = valueIndex as LogoNumber;
			int intIndex = (int)index.ToInt();

			// Out of bound
			string stringValue = Show();
			if ( intIndex < 1 || intIndex > stringValue.Length )
				throw new Exception("illegal value " + valueIndex.ToString());

			// Return char at position N
			return new LogoWord(Convert.ToString(stringValue[intIndex-1]));
		}

		/// <summary>
		/// Get the first item.
		/// </summary>
		/// <returns>First item</returns>
		public ILogoValue First()
		{
			// Return first char
			string stringValue = Show();
			return new LogoWord(Convert.ToString(stringValue[0]));
		}

		/// <summary>
		/// Get the last item.
		/// </summary>
		/// <returns>Last item</returns>
		public ILogoValue Last()
		{
			// Return last char
			string stringValue = Show();
			int length = stringValue.Length;
			return new LogoWord(Convert.ToString(stringValue[length-1]));
		}

		/// <summary>
		/// Get everything except the first item.
		/// </summary>
		/// <returns>Everything except the first item</returns>
		public ILogoValue ButFirst()
		{
			// Return everything but first char
			string stringValue = Show();
			int length = stringValue.Length;
			return new LogoWord(stringValue.Substring(1));
		}

		/// <summary>
		/// Get everything except the last item.
		/// </summary>
		/// <returns>Everything except the last item</returns>
		public ILogoValue ButLast()
		{
			// Return everything but last char
			string stringValue = Show();
			int length = stringValue.Length;
			return new LogoWord(stringValue.Substring(0, length-1));
		}

		/// <summary>
		/// Put a new item at the first position.
		/// </summary>
		/// <param name="valueToPut">New item to put</param>
		/// <returns>A new value including the new item</returns>
		public ILogoValue FPut(ILogoValue valueToPut)
		{
			throw new NotSupportedException("FPut on Number");
		}

		/// <summary>
		/// Put a new item at the last position.
		/// </summary>
		/// <param name="toPut">New item to put</param>
		/// <returns>A new value including the new item</returns>
		public ILogoValue LPut(ILogoValue toPut)
		{
			throw new NotSupportedException("LPut on Number");
		}

		/// <summary>
		/// Get the length of the value.
		/// </summary>
		/// <returns>Length of the value</returns>
		public int Length()
		{
			// Return number of char for value
			string stringValue = Show();
			return stringValue.Length;
		}
		#endregion

		#region Numeric Operations
		/// <summary>
		/// Add a value to the current value.
		/// </summary>
		/// <param name="valueToAdd">Value to add</param>
		/// <returns>The result of sum</returns>
		public ILogoValue Sum(ILogoValue valueToAdd)
		{
			// Value can't be a List
			if ( valueToAdd.IsList() || valueToAdd.IsArray() )
                throw new Exception(valueToAdd.ToString() + " must be a number");

			// Simplest case: two number
            LogoNumber valueToAddNumber;
            if (valueToAdd.IsNumber())
            {
                valueToAddNumber = valueToAdd as LogoNumber;
                return new LogoNumber(currentValue + valueToAddNumber.currentValue);
            }

            // Try to convert word to number
            valueToAddNumber = CreateFromWord(valueToAdd as LogoWord);
            if (valueToAddNumber == null)
            {
				// Convert to string
				LogoWord thisToWord = new LogoWord(this);
				return thisToWord.Sum(valueToAdd as LogoWord);
            }

            // Add number and number
            return new LogoNumber(currentValue + valueToAddNumber.currentValue);
		}

		/// <summary>
		/// Substract a value to the current value
		/// </summary>
		/// <param name="valueToDiff">Value to substract</param>
		/// <returns>The result of difference</returns>
		public ILogoValue Difference(ILogoValue valueToDiff)
		{
			// Value can't be a List
			if ( valueToDiff.IsList() )
                throw new Exception(valueToDiff.ToString() + " must be a number");

            // Simplest case: two number
            LogoNumber valueToDiffNumber;
            if (valueToDiff.IsNumber())
            {
                valueToDiffNumber = valueToDiff as LogoNumber;
                return new LogoNumber(currentValue - valueToDiffNumber.currentValue);
            }

            // Try to convert word to number
            valueToDiffNumber = CreateFromWord(valueToDiff as LogoWord);
            if (valueToDiffNumber == null)
                throw new Exception(valueToDiff.ToString() + " must be number");

            return new LogoNumber(currentValue - valueToDiffNumber.currentValue);
		}

		/// <summary>
		/// Multiply a value by the current value
		/// </summary>
		/// <param name="valueToMult">Value to multiply</param>
		/// <returns>The result of product</returns>
		public ILogoValue Product(ILogoValue valueToMult)
		{
			// Value can't be a List
			if ( valueToMult.IsList() || valueToMult.IsArray() )
                throw new Exception(valueToMult.ToString() + " must be a number");

            // Simplest case: two number
            LogoNumber valueToMultNumber;
            if (valueToMult.IsNumber())
            {
                valueToMultNumber = valueToMult as LogoNumber;
                return new LogoNumber(currentValue * valueToMultNumber.currentValue);
            }

            // Try to convert word to number
            valueToMultNumber = CreateFromWord(valueToMult as LogoWord);
            if (valueToMultNumber == null)
                throw new Exception(valueToMult.ToString() + " must be a number");

            return new LogoNumber(currentValue * valueToMultNumber.currentValue);
		}

		/// <summary>
		/// Divide a value by the current value
		/// </summary>
		/// <param name="valueToDiv">Value to divide</param>
		/// <returns>The result of quotient</returns>
		public ILogoValue Quotient(ILogoValue valueToDiv)
		{
			// Value can't be a List
			if ( valueToDiv.IsList() || valueToDiv.IsArray() )
                throw new Exception(valueToDiv.ToString() + " must be numberType");

            // Simplest case: two number
            LogoNumber valueToDivNumber;
            if (valueToDiv.IsNumber())
            {
                valueToDivNumber = valueToDiv as LogoNumber;
                return new LogoNumber(currentValue / valueToDivNumber.currentValue);
            }

            // Try to convert word to number
            valueToDivNumber = CreateFromWord(valueToDiv as LogoWord);
            if (valueToDivNumber == null)
                throw new Exception(valueToDiv.ToString() + "must be a number");

            return new LogoNumber(currentValue / valueToDivNumber.currentValue);
		}

		/// <summary>
		/// Change sign of the current value.
		/// </summary>
		/// <returns>Negative of the value</returns>
		public ILogoValue Minus()
		{
			// Return negative value
			return new LogoNumber(-currentValue);
		}
		#endregion

		#region Comparison Operations
		/// <summary>
		/// Test if current value is equal to value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if values are equal</returns>
		public bool IsEqual(ILogoValue valueToCompare)
		{
			// Value can't be a List
			if ( valueToCompare.IsList() || valueToCompare.IsArray() )
                throw new Exception(valueToCompare.ToString() + " must be a number");

            // Simplest case: two number
            LogoNumber valueAsNumber;
            if (valueToCompare.IsNumber())
            {
                valueAsNumber = valueToCompare as LogoNumber;
                return currentValue == valueAsNumber.currentValue;
            }

            // Try to convert word to number
            valueAsNumber = CreateFromWord(valueToCompare as LogoWord);
            if (valueAsNumber != null)
                return currentValue == valueAsNumber.currentValue;

			// Convert to word
			LogoWord thisToWord = new LogoWord(this);
			return thisToWord.IsEqual(valueToCompare);
		}

		/// <summary>
		/// Test if current value is not equal to value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if values are not equal</returns>
		public bool NotEqual(ILogoValue valueToCompare)
		{
            // Value can't be a List
            if (valueToCompare.IsList() || valueToCompare.IsArray())
                throw new Exception(valueToCompare.ToString() + " must be a number");

            // Simplest case: two number
            LogoNumber valueAsNumber;
            if (valueToCompare.IsNumber())
            {
                valueAsNumber = valueToCompare as LogoNumber;
                return currentValue != valueAsNumber.currentValue;
            }

            // Try to convert word to number
            valueAsNumber = CreateFromWord(valueToCompare as LogoWord);
            if (valueAsNumber != null)
                return currentValue != valueAsNumber.currentValue;

            // Convert to word
            LogoWord thisToWord = new LogoWord(this);
            return thisToWord.NotEqual(valueToCompare);
		}

		/// <summary>
		/// Test if current value is greater than value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if current value is greater than value</returns>
		public bool IsGreater(ILogoValue valueToCompare)
		{
            // Value can't be a List
            if (valueToCompare.IsList() || valueToCompare.IsArray())
                throw new Exception(valueToCompare.ToString() + " must be a number");

            // Simplest case: two number
            LogoNumber valueAsNumber;
            if (valueToCompare.IsNumber())
            {
                valueAsNumber = valueToCompare as LogoNumber;
                return currentValue > valueAsNumber.currentValue;
            }

            // Try to convert word to number
            valueAsNumber = CreateFromWord(valueToCompare as LogoWord);
            if (valueAsNumber != null)
                return currentValue > valueAsNumber.currentValue;

            // Convert to word
            LogoWord thisToWord = new LogoWord(this);
            return thisToWord.IsGreater(valueToCompare);
		}


		/// <summary>
		/// Test if current value is lower than value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if current value is lower than value</returns>
		public bool IsLower(ILogoValue valueToCompare)
		{
            // Value can't be a List
            if (valueToCompare.IsList() || valueToCompare.IsArray())
                throw new Exception(valueToCompare.ToString() + " must be a number");

            // Simplest case: two number
            LogoNumber valueAsNumber;
            if (valueToCompare.IsNumber())
            {
                valueAsNumber = valueToCompare as LogoNumber;
                return currentValue < valueAsNumber.currentValue;
            }

            // Try to convert word to number
            valueAsNumber = CreateFromWord(valueToCompare as LogoWord);
            if (valueAsNumber != null)
                return currentValue < valueAsNumber.currentValue;

            // Convert to word
            LogoWord thisToWord = new LogoWord(this);
            return thisToWord.IsLower(valueToCompare);
		}

		/// <summary>
		/// Test if current value is greater or equal to value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if current value is greater or equal to value</returns>
		public bool IsGreaterOrEqual(ILogoValue valueToCompare)
		{
            // Value can't be a List
            if (valueToCompare.IsList() || valueToCompare.IsArray())
                throw new Exception(valueToCompare.ToString() + " must be a number");

            // Simplest case: two number
            LogoNumber valueAsNumber;
            if (valueToCompare.IsNumber())
            {
                valueAsNumber = valueToCompare as LogoNumber;
                return currentValue >= valueAsNumber.currentValue;
            }

            // Try to convert word to number
            valueAsNumber = CreateFromWord(valueToCompare as LogoWord);
            if (valueAsNumber != null)
                return currentValue >= valueAsNumber.currentValue;

            // Convert to word
            LogoWord thisToWord = new LogoWord(this);
            return thisToWord.IsGreaterOrEqual(valueToCompare);
		}

		/// <summary>
		/// Test if current value is lower or equal to value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if current value is lower or equal to value</returns>
		public bool IsLowerOrEqual(ILogoValue valueToCompare)
		{
            // Value can't be a List
            if (valueToCompare.IsList() || valueToCompare.IsArray())
                throw new Exception(valueToCompare.ToString() + " must be a number");

            // Simplest case: two number
            LogoNumber valueAsNumber;
            if (valueToCompare.IsNumber())
            {
                valueAsNumber = valueToCompare as LogoNumber;
                return currentValue <= valueAsNumber.currentValue;
            }

            // Try to convert word to number
            valueAsNumber = CreateFromWord(valueToCompare as LogoWord);
            if (valueAsNumber != null)
                return currentValue <= valueAsNumber.currentValue;

            // Convert to word
            LogoWord thisToWord = new LogoWord(this);
            return thisToWord.IsLowerOrEqual(valueToCompare);
		}
		#endregion

		#region Other Operations
		/// <summary>
		/// Do a copy of the value.
		/// </summary>
		/// <returns>a strict copy of the current value</returns>
		public ILogoValue Copy()
		{
			return new LogoNumber(currentValue);
		}

		/// <summary>
		/// Convert the current value to a printable string.
		/// </summary>
		/// <returns>a string with the value</returns>
		public string Print()
		{
			return currentValue.ToString();
		}

		/// <summary>
		/// Convert the current value to a showable string.
		/// </summary>
		/// <returns>a string with the value</returns>
		public string Show()
		{
			// Use a neutral format for separators
            return currentValue.ToString(numberProvider);
		}

		/// <summary>
		/// Convert to an int value
		/// </summary>
		/// <returns>int</returns>
		public int ToInt()
		{
			return (int)currentValue;
		}

		/// <summary>
		/// Convert to a double value
		/// </summary>
		/// <returns>double</returns>
		public double ToDouble()
		{
			return currentValue;
		}

		/// <summary>
		/// Convert the current value to a printable string.
		/// </summary>
		/// <returns>a string with the value</returns>
		public override string ToString()
		{
			// Use a neutral format for separators
            return currentValue.ToString(numberProvider);
		}
		#endregion
        
        #region ICallable interface
        /// <summary>
        /// Call a method on the object.
        /// </summary>
        /// <param name="name">method name</param>
        /// <param name="args">parameters</param>
        /// <returns>return value for the method</returns>
        public object CallMethod(string name, params object[] args)
        {
            name = name.ToLower();
            int count = args.Length;
            if ((name.CompareTo("butfirst") == 0 || name.CompareTo("bf") == 0) && count == 0)
                return ButFirst();
            else if ((name.CompareTo("butlast") == 0 || name.CompareTo("bl") == 0) && count == 0)
                return ButLast();
            if (name.CompareTo("first") == 0 && count == 0)
                return First();
            else if (name.CompareTo("last") == 0 && count == 0)
                return Last();
            else if (name.CompareTo("item") == 0 && count == 1)
                return Item((ILogoValue)(args[0].ConvertTo(typeof(ILogoValue))));
            else if (name.CompareTo("fput") == 0 && count == 1)
                return FPut((ILogoValue)(args[0].ConvertTo(typeof(ILogoValue))));

            throw new NotSupportedException(name);
        }

        /// <summary>
        /// Get a property value on the object.
        /// </summary>
        /// <param name="name">property name</param>
        /// <returns>property value</returns>
        public object GetProperty(string name)
        {
            name = name.ToLower();
            if (name.CompareTo("count") == 0)
                return new LogoNumber(Length());

            throw new NotSupportedException();
        }

        /// <summary>
        /// Set value for a property on the object.
        /// </summary>
        /// <param name="name">property name</param>
        /// <param name="value">new property value</param>
        public void SetProperty(string name, object value)
        {
            throw new NotSupportedException();
        }
        #endregion
	}
}
