#region LIOGO - GPL Copyright (c) 2005-2007 Lionel Lask
//
// This file is part of LIOGO.
//
// LIOGO is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// LIOGO is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Liogo; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//================================================================================
#endregion

using System;
using System.ComponentModel;
using System.Globalization;
using MyDLR;

namespace LIOGO.Core
{
	/// <summary>
	/// Class for a Logo Word.
	/// </summary>
    [TypeConverter(typeof(LogoConverter))]   // Added for MyDLR project
	public class LogoWord : ILogoValue
	{
		/// <summary>
		/// Current value for the Word.
		/// </summary>
		private string currentValue;

		/// <summary>
		/// This word has quote.
		/// </summary>
		private bool isQuoted;

        /// <summary>
        /// Ignore case on comparaison
        /// </summary>
        private const bool ignoreCase = false;
		
		#region Constructors
		/// <summary>
		/// Constructor with nothing.
		/// </summary>
		public LogoWord()
		{
			currentValue = null;
			isQuoted = true;
		}

		/// <summary>
		/// Constructor with a string.
		/// </summary>
		/// <param name="initValue">initial value</param>
		public LogoWord(string initValue)
		{
			currentValue = initValue;
			isQuoted = true;
		}

		/// <summary>
		/// Constructor with a string and with or without quote.
		/// </summary>
		/// <param name="initValue">initial value</param>
		/// <param name="isQuotedValue">word has quote or not</param>
		public LogoWord(string initValue, bool isQuotedValue)
		{
			currentValue = initValue;
			isQuoted = isQuotedValue;
		}

		/// <summary>
		/// Constructor with a bool: map to Logo true and false.
		/// </summary>
		/// <param name="initValue">initial value</param>
		public LogoWord(bool initValue)
		{
			if ( initValue )
				currentValue = LogoWord.True.Print();
			else
				currentValue = LogoWord.False.Print();
			isQuoted = true;
		}

		/// <summary>
		/// Constructor with a Logo value.
		/// </summary>
		/// <param name="initValue">initial value</param>
		public LogoWord(ILogoValue initValue)
		{
			// Can't convert from List
            if (initValue.IsList() || initValue.IsArray())
                throw new Exception(initValue.ToString() + " must be a word");

			// From Word to Word
			if (initValue.IsWord())
			{
				currentValue = (initValue as LogoWord).currentValue;
				isQuoted = true;
				return;
			}

			// From Number to Word
			currentValue = initValue.Show();
			isQuoted = true;
		}
		#endregion

		#region Test Operations
		/// <summary>
		/// Test if the value is a Word
		/// </summary>
		/// <returns>true if the value is a Word</returns>
		public bool IsWord()
		{
			return true;
		}

		/// <summary>
		/// Test if the value is a Number
		/// </summary>
		/// <returns>true if the value is a Number</returns>
		public bool IsNumber()
		{
			return false;
		}

		/// <summary>
		/// Test if the value is a List
		/// </summary>
		/// <returns>true if the value is a List</returns>
		public bool IsList()
		{
			return false;
		}

        /// <summary>
        /// Test if the value is an Array
        /// </summary>
        /// <returns>true if the value is an Array</returns>
        public bool IsArray()
        {
            return false;
        }

		/// <summary>
		/// Test if the value is empty
		/// </summary>
		/// <returns>true if the value is empty</returns>
		public bool IsEmpty()
		{
			return (currentValue == null || currentValue.Length == 0);
		}

		/// <summary>
		/// Test if the value is true
		/// </summary>
		/// <returns>true if the value is equal to true</returns>
		public bool IsTrue()
		{
			// Use internal value to avoid case effect
			return String.Compare(currentValue, trueValue.currentValue, true) == 0;
		}

		/// <summary>
		/// Test if the value is false
		/// </summary>
		/// <returns>true if the value is equal to false</returns>
		public bool IsFalse()
		{
			// Use internal value to avoid case effect
			return String.Compare(currentValue, falseValue.currentValue, true) == 0;
		}
		#endregion

		#region Composite Operations
		/// <summary>
		/// Get item N in the value.
		/// </summary>
		/// <param name="valueIndex">Index of item (start at 1)</param>
		/// <returns>Value at index</returns>
		public ILogoValue Item(ILogoValue valueIndex)
		{
			// Index must be a number or must be convertible to number
			LogoNumber index;
			if ( !valueIndex.IsNumber() )
			{
                if (valueIndex.IsList() || valueIndex.IsArray())
                    throw new Exception(valueIndex.ToString() + " must be a word");
				index = new LogoNumber(valueIndex);
			}
			else
				index = valueIndex as LogoNumber;
			int intIndex = (int)index.ToInt();

			// Out of bound
			if ( currentValue == null || intIndex < 1 || intIndex > currentValue.Length )
				throw new Exception("illegal value "+valueIndex.ToString());

			// Return char at position N
			return new LogoWord(Convert.ToString(currentValue[intIndex-1]));
		}

		/// <summary>
		/// Get the first item.
		/// </summary>
		/// <returns>First item</returns>
		public ILogoValue First()
		{
			// Empty string
			if ( currentValue == null || currentValue.Length == 0 )
				return new LogoWord();

			// Return first char
			return new LogoWord(Convert.ToString(currentValue[0]));	
		}

		/// <summary>
		/// Get the last item.
		/// </summary>
		/// <returns>Last item</returns>
		public ILogoValue Last()
		{
			// Empty string
			if ( currentValue == null || currentValue.Length == 0 )
				return new LogoWord();

			// Return last char
			return new LogoWord(Convert.ToString(currentValue[currentValue.Length-1]));
		}

		/// <summary>
		/// Get everything except the first item.
		/// </summary>
		/// <returns>Everything except the first item</returns>
		public ILogoValue ButFirst()
		{
			// Empty string
			if ( currentValue == null || currentValue.Length == 0 )
				return new LogoWord();

			// Return last char
			return new LogoWord(currentValue.Substring(1));
		}

		/// <summary>
		/// Get everything except the last item.
		/// </summary>
		/// <returns>Everything except the last item</returns>
		public ILogoValue ButLast()
		{
			// Empty string
			if ( currentValue == null || currentValue.Length == 0 )
				return new LogoWord();

			// Return last char
			return new LogoWord(currentValue.Substring(0, currentValue.Length-1));
		}

		/// <summary>
		/// Put a new item at the first position.
		/// </summary>
		/// <param name="valueToPut">New item to put</param>
		/// <returns>A new value including the new item</returns>
		public ILogoValue FPut(ILogoValue valueToPut)
		{
			throw new NotSupportedException("FPut on Word");
		}

		/// <summary>
		/// Put a new item at the last position.
		/// </summary>
		/// <param name="toPut">New item to put</param>
		/// <returns>A new value including the new item</returns>
		public ILogoValue LPut(ILogoValue toPut)
		{
			throw new NotSupportedException("LPut on Word");
		}

		/// <summary>
		/// Get the length of the value.
		/// </summary>
		/// <returns>Length of the value</returns>
		public int Length()
		{
			// Empty value
			if ( currentValue == null )
				return 0;

			return currentValue.Length;
		}
		#endregion

		#region Numeric Operations
		/// <summary>
		/// Add a value to the current value.
		/// </summary>
		/// <param name="valueToAdd">Value to add</param>
		/// <returns>The result of sum</returns>
		public ILogoValue Sum(ILogoValue valueToAdd)
		{
			// Value can't be a List
			if ( valueToAdd.IsList() || valueToAdd.IsArray() )
                throw new Exception(valueToAdd.ToString() + " must be a word");

			// Try number first
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber != null)
            {
                LogoNumber valueToProcessToNumber;
                if (valueToAdd.IsNumber())
                {
                    valueToProcessToNumber = (LogoNumber)valueToAdd;
                    return thisToNumber.Sum(valueToProcessToNumber); 
                }
                else
                    valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)valueToAdd);
                if (valueToProcessToNumber != null)
                    return thisToNumber.Sum(valueToProcessToNumber); 
            }

			// Add an empty value
			LogoWord toAdd = new LogoWord(valueToAdd);
			string toAddValue = toAdd.currentValue;
			if (toAddValue == null || toAddValue.Length == 0 )
				return this;

			// Return concatenated strings
			return new LogoWord(currentValue+toAddValue);
		}

		/// <summary>
		/// Substract a value to the current value
		/// </summary>
		/// <param name="toValue">Value to substract</param>
		/// <returns>The result of difference</returns>
		public ILogoValue Difference(ILogoValue toValue)
		{
            // Try number
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber == null || toValue.IsList() || toValue.IsArray())
                throw new NotSupportedException("Difference on Word");
            LogoNumber valueToProcessToNumber;
            if (toValue.IsNumber())
                valueToProcessToNumber = (LogoNumber)toValue;
            else
            {
                valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)toValue);
                if (valueToProcessToNumber == null)
                    throw new NotSupportedException("Difference on Word");
            }
            return thisToNumber.Difference(valueToProcessToNumber);
		}

		/// <summary>
		/// Multiply a value by the current value
		/// </summary>
		/// <param name="toValue">Value to multiply</param>
		/// <returns>The result of product</returns>
		public ILogoValue Product(ILogoValue toValue)
		{
            // Try number
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber == null || toValue.IsList() || toValue.IsArray())
                throw new NotSupportedException("Product on Word");
            LogoNumber valueToProcessToNumber;
            if (toValue.IsNumber())
                valueToProcessToNumber = (LogoNumber)toValue;
            else
            {
                valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)toValue);
                if (valueToProcessToNumber == null)
                    throw new NotSupportedException("Product on Word");
            }
            return thisToNumber.Product(valueToProcessToNumber);
		}

		/// <summary>
		/// Divide a value by the current value
		/// </summary>
		/// <param name="toValue">Value to divide</param>
		/// <returns>The result of quotient</returns>
		public ILogoValue Quotient(ILogoValue toValue)
		{
            // Try number
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber == null || toValue.IsList() || toValue.IsArray())
                throw new NotSupportedException("Quotient on Word");
            LogoNumber valueToProcessToNumber;
            if (toValue.IsNumber())
                valueToProcessToNumber = (LogoNumber)toValue;
            else
            {
                valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)toValue);
                if (valueToProcessToNumber == null)
                    throw new NotSupportedException("Quotient on Word");
            }
            return thisToNumber.Quotient(valueToProcessToNumber);
		}

		/// <summary>
		/// Change sign of the current value.
		/// </summary>
		/// <returns>Negative of the value</returns>
		public ILogoValue Minus()
		{
            // Try number
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber == null)
                throw new NotSupportedException("Minus on Word");
            return thisToNumber.Minus();
		}
		#endregion

		#region Comparison Operations
		/// <summary>
		/// Test if current value is equal to value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if values are equal</returns>
		public bool IsEqual(ILogoValue valueToCompare)
		{
			// Value can't be a List or an Array
			if ( valueToCompare.IsList() || valueToCompare.IsArray() )
				return false;

            // Try number first
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber != null)
            {
                LogoNumber valueToProcessToNumber;
                if (valueToCompare.IsNumber())
                {
                    valueToProcessToNumber = (LogoNumber)valueToCompare;
                    return thisToNumber.IsEqual(valueToProcessToNumber);
                }
                else
                    valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)valueToCompare);
                if (valueToProcessToNumber != null)
                    return thisToNumber.IsEqual(valueToProcessToNumber);
            }

			// Compare string
			LogoWord toCompare = new LogoWord(valueToCompare);
			string toCompareValue = toCompare.currentValue;
			return String.Compare(currentValue, toCompareValue, ignoreCase) == 0;
		}

		/// <summary>
		/// Test if current value is not equal to value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if values are not equal</returns>
		public bool NotEqual(ILogoValue valueToCompare)
		{
			// Value can't be a List or an Array
			if ( valueToCompare.IsList() || valueToCompare.IsArray() )
				return true;

            // Try number first
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber != null)
            {
                LogoNumber valueToProcessToNumber;
                if (valueToCompare.IsNumber())
                {
                    valueToProcessToNumber = (LogoNumber)valueToCompare;
                    return thisToNumber.NotEqual(valueToProcessToNumber);
                }
                else
                    valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)valueToCompare);
                if (valueToProcessToNumber != null)
                    return thisToNumber.NotEqual(valueToProcessToNumber);
            }

			// Compare string
			LogoWord toCompare = new LogoWord(valueToCompare);
			string toCompareValue = toCompare.currentValue;
			return String.Compare(currentValue, toCompareValue, ignoreCase) != 0;
		}

		/// <summary>
		/// Test if current value is greater than value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if current value is greater than value</returns>
		public bool IsGreater(ILogoValue valueToCompare)
		{
			// Value can't be a List or an Array
			if ( valueToCompare.IsList() || valueToCompare.IsArray() )
                throw new Exception(valueToCompare.ToString() + " must be a word");

            // Try number first
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber != null)
            {
                LogoNumber valueToProcessToNumber;
                if (valueToCompare.IsNumber())
                {
                    valueToProcessToNumber = (LogoNumber)valueToCompare;
                    return thisToNumber.IsGreater(valueToProcessToNumber);
                }
                else
                    valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)valueToCompare);
                if (valueToProcessToNumber != null)
                    return thisToNumber.IsGreater(valueToProcessToNumber);
            }

			// Compare string
			LogoWord toCompare = new LogoWord(valueToCompare);
			string toCompareValue = toCompare.currentValue;
			return String.Compare(currentValue, toCompareValue, ignoreCase) > 0;
		}


		/// <summary>
		/// Test if current value is lower than value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if current value is lower than value</returns>
		public bool IsLower(ILogoValue valueToCompare)
		{
			// Value can't be a List or an Array
			if ( valueToCompare.IsList() || valueToCompare.IsArray() )
                throw new Exception(valueToCompare.ToString() + " must be a word");

            // Try number first
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber != null)
            {
                LogoNumber valueToProcessToNumber;
                if (valueToCompare.IsNumber())
                {
                    valueToProcessToNumber = (LogoNumber)valueToCompare;
                    return thisToNumber.IsLower(valueToProcessToNumber);
                }
                else
                    valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)valueToCompare);
                if (valueToProcessToNumber != null)
                    return thisToNumber.IsLower(valueToProcessToNumber);
            }

			// Compare string
			LogoWord toCompare = new LogoWord(valueToCompare);
			string toCompareValue = toCompare.currentValue;
			return String.Compare(currentValue, toCompareValue, ignoreCase) < 0;
		}

		/// <summary>
		/// Test if current value is greater or equal to value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if current value is greater or equal to value</returns>
		public bool IsGreaterOrEqual(ILogoValue valueToCompare)
		{
			// Value can't be a List or an Array
			if ( valueToCompare.IsList() || valueToCompare.IsArray() )
                throw new Exception(valueToCompare.ToString() + " must be a word");

            // Try number first
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber != null)
            {
                LogoNumber valueToProcessToNumber;
                if (valueToCompare.IsNumber())
                {
                    valueToProcessToNumber = (LogoNumber)valueToCompare;
                    return thisToNumber.IsGreaterOrEqual(valueToProcessToNumber);
                }
                else
                    valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)valueToCompare);
                if (valueToProcessToNumber != null)
                    return thisToNumber.IsGreaterOrEqual(valueToProcessToNumber);
            }

			// Compare string
			LogoWord toCompare = new LogoWord(valueToCompare);
			string toCompareValue = toCompare.currentValue;
			return String.Compare(currentValue, toCompareValue, ignoreCase) >= 0;
		}

		/// <summary>
		/// Test if current value is lower or equal to value.
		/// </summary>
		/// <param name="valueToCompare">value to compare</param>
		/// <returns>true if current value is lower or equal to value</returns>
		public bool IsLowerOrEqual(ILogoValue valueToCompare)
		{
			// Value can't be a List or an Array
			if ( valueToCompare.IsList() || valueToCompare.IsArray() )
                throw new Exception(valueToCompare.ToString() + " must be a word");

            // Try number first
            LogoNumber thisToNumber = LogoNumber.CreateFromWord(this);
            if (thisToNumber != null)
            {
                LogoNumber valueToProcessToNumber;
                if (valueToCompare.IsNumber())
                {
                    valueToProcessToNumber = (LogoNumber)valueToCompare;
                    return thisToNumber.IsLowerOrEqual(valueToProcessToNumber);
                }
                else
                    valueToProcessToNumber = LogoNumber.CreateFromWord((LogoWord)valueToCompare);
                if (valueToProcessToNumber != null)
                    return thisToNumber.IsLowerOrEqual(valueToProcessToNumber);
            }

			// Compare string
			LogoWord toCompare = new LogoWord(valueToCompare);
			string toCompareValue = toCompare.currentValue;
			return String.Compare(currentValue, toCompareValue, ignoreCase) <= 0;
		}
		#endregion

		#region Other Operations
		/// <summary>
		/// Do a copy of the value.
		/// </summary>
		/// <returns>a strict copy of the current value</returns>
		public ILogoValue Copy()
		{
			return new LogoWord(currentValue, isQuoted);
		}

		/// <summary>
		/// Do a copy of the value with or without quote.
		/// </summary>
		/// <param name="withQuote">use quote or not</param>
		/// <returns>a strict copy of the current value</returns>
		public ILogoValue Copy(bool withQuote)
		{
			return new LogoWord(currentValue, withQuote);
		}

		/// <summary>
		/// Convert the current value to a printable string.
		/// </summary>
		/// <returns>a string with the value</returns>
		public string Print()
		{
			if ( currentValue == null )
				return "";
			return currentValue;
		}

		/// <summary>
		/// Convert the current value to a showable string.
		/// </summary>
		/// <returns>a string with the value</returns>
		public string Show()
		{
			if ( currentValue == null )
				return "";
			return currentValue;
		}

		/// <summary>
		/// Convert the current value to the initially parsed string.
		/// </summary>
		/// <returns>a string with the value</returns>
		public override string ToString()
		{
			if ( isQuoted )
			{
				if ( currentValue == null )
					return "\"";
				if ( currentValue.IndexOf(" ") != -1 || currentValue.IndexOf("\t") != -1 || currentValue.IndexOf("\r\n") != -1 )
					return "\"|"+currentValue+"|";
				return "\""+currentValue;
			}
			return Print();
		}
		#endregion

		#region Boolean constant
		/// <summary>
		/// Static initialisation
		/// </summary>
		static LogoWord()
		{
            trueValue = new LogoWord("true");
            falseValue = new LogoWord("false");
		}

		/// <summary>
		/// True
		/// </summary>
		static public ILogoValue True
		{
			get { return trueValue; }
		}
		static private LogoWord trueValue;

		/// <summary>
		/// False
		/// </summary>
		static public ILogoValue False
		{
			get { return falseValue; }
		}
		static private LogoWord falseValue;
		#endregion

        #region ICallable interface
        /// <summary>
        /// Call a method on the object.
        /// </summary>
        /// <param name="name">method name</param>
        /// <param name="args">parameters</param>
        /// <returns>return value for the method</returns>
        public object CallMethod(string name, params object[] args)
        {
            name = name.ToLower();
            int count = args.Length;
            if ((name.CompareTo("butfirst") == 0 || name.CompareTo("bf") == 0)&& count == 0)
                return ButFirst();
            else if ((name.CompareTo("butlast") == 0 || name.CompareTo("bl") == 0)&& count == 0)
                return ButLast();
            if (name.CompareTo("first") == 0 && count == 0)
                return First();
            else if (name.CompareTo("last") == 0 && count == 0)
                return Last();
            else if (name.CompareTo("item") == 0 && count == 1)
                return Item((ILogoValue)(args[0].ConvertTo(typeof(ILogoValue))));
            else if (name.CompareTo("fput") == 0 && count == 1)
                return FPut((ILogoValue)(args[0].ConvertTo(typeof(ILogoValue)))); 

            throw new NotSupportedException(name);
        }

        /// <summary>
        /// Get a property value on the object.
        /// </summary>
        /// <param name="name">property name</param>
        /// <returns>property value</returns>
        public object GetProperty(string name)
        {
            name = name.ToLower();
            if (name.CompareTo("count") == 0)
                return new LogoNumber(Length());

            throw new NotSupportedException();
        }

        /// <summary>
        /// Set value for a property on the object.
        /// </summary>
        /// <param name="name">property name</param>
        /// <param name="value">new property value</param>
        public void SetProperty(string name, object value)
        {
            throw new NotSupportedException();
        }
        #endregion
    }
}
