﻿using System;
using System.ComponentModel;
using System.Reflection;

namespace MyDLR
{
    /// <summary>
    /// Extension to handle dynamic methods on CLR object using Reflection.
    /// </summary>
    public static class DynamicObjectExtension
    {
        #region Extension equivalent for ICallable interface
        /// <summary>
        /// Call a method on the object.
        /// </summary>
        /// <param name="name">method name</param>
        /// <param name="args">parameters</param>
        /// <returns>return value for the method</returns>
        public static object CallMethod(this object @this, string name, params object[] args)
        {
            // Try dynamic object interface first
            if (@this is ICallable)
            {
                ICallable dynthis = @this as ICallable;
                return dynthis.CallMethod(name, args);
            }

            // Look each methods 
            Type myType = @this.GetType();
            MethodInfo[] methods = myType.GetMethods();
            int length = methods.Length;
            for (int i = 0; i < length; i++)
            {
                // If method match
                MethodInfo current = methods[i];
                ParameterInfo[] argsinfo = current.GetParameters();
                if (name.ToLower() == current.Name.ToLower()
                    && current.IsPublic
                    && !current.IsStatic
                    && !current.IsSpecialName
                    && args.Length == argsinfo.Length)
                {
                    // Try to convert each args to the right destination
                    bool typeMatch = true;
                    for(int j = 0; j < argsinfo.Length && typeMatch; j++)
                    {
                        ParameterInfo arginfo = argsinfo[j];
                        try
                        {
                            args[j] = args[j].ConvertTo(arginfo.ParameterType);
                        }
                        catch (NotSupportedException)
                        {
                            typeMatch = false;
                        }
                    }
                    if (!typeMatch)
                        continue;

                    // Invoke it
                    return current.Invoke(@this, args);
                }
            }

            // Not found
            throw new NotSupportedException(name);
        }

        /// <summary>
        /// Get a property value on the object.
        /// </summary>
        /// <param name="name">property name</param>
        /// <returns>property value</returns>
        public static object GetProperty(this object @this, string name)
        {
            // Try dynamic object interface first
            if (@this is ICallable)
            {
                ICallable dynthis = @this as ICallable;
                return dynthis.GetProperty(name);
            }

            // Look each properties 
            Type myType = @this.GetType();
            PropertyInfo[] properties = myType.GetProperties();
            int length = properties.Length;
            for (int i = 0; i < length; i++)
            {
                // If method match
                PropertyInfo current = properties[i];
                if (name.ToLower() == current.Name.ToLower()
                    && current.CanRead
                    && !current.IsSpecialName)
                {
                    return current.GetValue(@this, null);
                }
            }

            // Not found
            throw new NotSupportedException();
        }

        /// <summary>
        /// Set value for a property on the object.
        /// </summary>
        /// <param name="name">property name</param>
        /// <param name="value">new property value</param>
        public static void SetProperty(this object @this, string name, object value)
        {
            // Try dynamic object interface first
            if (@this is ICallable)
            {
                ICallable dynthis = @this as ICallable;
                dynthis.SetProperty(name, value);
                return;
            }

            // TODO: CLR.SetProperty
            throw new NotImplementedException();
        }
        #endregion

        /// <summary>
        /// Create an instance of the class specified by the string.
        /// </summary>
        /// <param name="classname">name of class</param>
        /// <param name="args">parameters for constructor</param>
        /// <returns>the new instance</returns>
        public static object CreateInstance(this string classname, params object[] args)
        {
            // Look in DLR languages first
            foreach (IDLRProvider dle in Providers.languages)
            {
                if (dle.HasClass(classname))
                    return dle.CreateInstance(classname, args);
            }

            // Look for a CLR destination
            Type type = Type.GetType(classname);
            if (type != null)
                return Activator.CreateInstance(type, args);

            throw new NotSupportedException(classname);
        }

        /// <summary>
        /// Convert the object to the specified type.
        /// </summary>
        /// <param name="destination">type expected</param>
        /// <returns>value in the new type or null</returns>
        public static object ConvertTo(this object @this, Type destination)
        {
            // Try to use standard conversion
            Type source = @this.GetType();
            if (destination.IsAssignableFrom(source))
                return @this;

            // Try CLR conversion using the source converter
            TypeConverter srcconverter = TypeDescriptor.GetConverter(source);
            if (srcconverter != null)
            {
                if (srcconverter.CanConvertTo(destination))
                    return srcconverter.ConvertTo(@this, destination);
            }

            // Try CLR conversion using the destination converter
            TypeConverter destconverter = TypeDescriptor.GetConverter(destination);
            if (destconverter != null)
            {
                if (destconverter.CanConvertFrom(source))
                    return destconverter.ConvertFrom(@this);
            }

            // If both type came from DLR, try to use string as destination in the middle
            if (@this is ICallable && Util.IsDLRType(destination))
            {
                if (srcconverter.CanConvertTo(typeof(string)) && destconverter.CanConvertFrom(typeof(string)))
                {
                    object asstring = srcconverter.ConvertTo(@this, typeof(string));
                    return destconverter.ConvertFrom(asstring);
                }
            }

            throw new NotSupportedException(destination.ToString());
        }
    }
}
