﻿using System;
using MyDLR;

namespace MyJScript
{
    /// <summary>
    /// DLR provider for MyJScript.
    /// </summary>
    public class MJSDLRProvider : IDLRProvider
    {
        /// <summary>
        /// Test is the language contain a definition for a class.
        /// </summary>
        /// <param name="classname">class name</param>
        /// <returns>true if language has this class</returns>
        public bool HasClass(string classname)
        {
            // Look for primitives class
            if (classname.Equals("String") || classname.Equals("Number"))
                return true;

            // Look in global context
            MJSObject o = MJSContext.GetContext().Get(classname);
            if (o == null)
                return false;

            // Must be a method name
            return (o is MJSFunction);
        }

        /// <summary>
        /// Create an instance of the class using parameters values.
        /// </summary>
        /// <param name="classname">class name</param>
        /// <param name="args">arguments for constructor</param>
        /// <returns>the newly created object</returns>
        public object CreateInstance(string classname, params object[] args)
        {
            // Is it a MJSScript class ?
            if (!HasClass(classname))
                throw new NotSupportedException(classname);

            // Primitive constructors first
            if (classname.Equals("String"))
            {
                if (args.Length != 1)
                    throw new NotSupportedException(classname);
                object param = args[0];
                Type type = param.GetType();
                if (type == typeof(string))
                    return new MJSString(param as string);
                else
                    return new MJSString(param.ConvertTo(typeof(string)) as string);
            }
            else if (classname.Equals("Number"))
            {
                if (args.Length != 1)
                    throw new NotSupportedException(classname);
                object param = args[0];
                Type type = param.GetType();
                if (type == typeof(Int32))
                    return new MJSNumber((Int32)param);
                else
                    return new MJSNumber((Int32)param.ConvertTo(typeof(Int32)));
            }

            // Convert each param to a MJSObject
            int length = args.Length;
            MJSObject[] mjsargs = new MJSObject[length];
            for (int i = 0; i < length; i++)
            {
                object param = args[i];
                Type type = param.GetType();
                if (type.IsSubclassOf(typeof(MJSObject)))
                    mjsargs[i] = param as MJSObject;
                else if (type == typeof(Int32))
                    mjsargs[i] = new MJSNumber((Int32)param);
                else if (type == typeof(string))
                    mjsargs[i] = new MJSString(param as string);
                else
                    mjsargs[i] = param.ConvertTo(typeof(MJSObject)) as MJSObject;
            }

            // Get constructor method
            MJSObject o = MJSContext.GetContext().Get(classname);
            MJSFunction constructor = o as MJSFunction;

            // Create a new object and call constructor on it
            MJSObject newobject = new MJSObject();
            constructor.Call(newobject, mjsargs);

            return newobject;
        }
    }
}
