﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using MyDLR;

namespace MyJScript
{
    /// <summary>
    /// Class for a MyJScript object.
    /// </summary>
    [TypeConverter(typeof(MJSConverter))]
    public class MJSObject : ICallable
    {
        /// <summary>
        ///  Variables instance and methods instance
        /// </summary>
        Dictionary<string, MJSObject> members;

        /// <summary>
        /// Constructor
        /// </summary>
        public MJSObject()
        {
            members = new Dictionary<string, MJSObject>();
        }

        /// <summary>
        /// Set a member value.
        /// </summary>
        /// <param name="name">member name</param>
        /// <param name="value">new value</param>
        public virtual void Set(string name, MJSObject value)
        {
            // Remove it first if need
            if (members.ContainsKey(name))
                members.Remove(name);

            // Add property
            members.Add(name, value);
        }

        /// <summary>
        /// Get a member value.
        /// </summary>
        /// <param name="name">member name</param>
        /// <returns></returns>
        public virtual MJSObject Get(string name)
        {
            if (!members.ContainsKey(name))
                return null;

            return members[name];
        }

        /// <summary>
        /// Get the value.
        /// </summary>
        /// <returns>value</returns>
        public virtual object ValueOf()
        {
            return members;
        }

        /// <summary>
        /// Convert to string.
        /// </summary>
        /// <returns>Object as string</returns>
        public override string ToString()
        {
            bool first = true;
            StringBuilder res = new StringBuilder("{");
            foreach (KeyValuePair<string, MJSObject> member in members)
            {
                if (!first)
                    res.Append(", ");
                res.Append('"');
                res.Append(member.Key);
                res.Append('"');
                res.Append(":");
                res.Append(member.Value.ToString());
                first = false;
            }
            res.Append("}");

            return res.ToString();
        }

        #region ICallable interface
        /// <summary>
        /// Call a method on the object.
        /// </summary>
        /// <param name="name">method name</param>
        /// <param name="args">parameters</param>
        /// <returns>return value for the method</returns>
        public object CallMethod(string name, params object[] args)
        {
            // Get property using this name
            MJSObject value = Get(name);
            if (value == null)
                throw new NotSupportedException(name);

            // Method must be only on MJSFunction
            MJSFunction method = value as MJSFunction;
            if (method == null)
                throw new NotSupportedException(name);

            // Convert each param to a MJSObject
            int length = args.Length;
            MJSObject[] mjsargs = new MJSObject[length];
            for (int i = 0; i < length; i++)
            {
                object param = args[i];
                Type type = param.GetType();
                if (type.IsSubclassOf(typeof(MJSObject)))
                    mjsargs[i] = param as MJSObject;
                else if (type == typeof(Int32))
                    mjsargs[i] = new MJSNumber((Int32)param);
                else if (type == typeof(string))
                    mjsargs[i] = new MJSString(param as string);
                else
                    mjsargs[i] = new MJSString(param.ConvertTo(typeof(string)) as string);
            }

            return method.Call(this, mjsargs);
        }

        /// <summary>
        /// Get a property value on the object.
        /// </summary>
        /// <param name="name">property name</param>
        /// <returns>property value</returns>
        public object GetProperty(string name)
        {
            return Get(name);
        }

        /// <summary>
        /// Set value for a property on the object.
        /// </summary>
        /// <param name="name">property name</param>
        /// <param name="value">new property value</param>
        public void SetProperty(string name, object value)
        {
            // Convert value to MJSObject
            MJSObject mjsvalue;
            Type type = value.GetType();
            if (type.IsSubclassOf(typeof(MJSObject)))
                mjsvalue = value as MJSObject;
            else if (type == typeof(Int32))
                mjsvalue = new MJSNumber((Int32)value);
            else if (type == typeof(string))
                mjsvalue = new MJSString(value as string);
            else
                mjsvalue = new MJSString(value.ConvertTo(typeof(string)) as string);
            
            // Set property
            Set(name, mjsvalue);
        }
        #endregion
    }
}
