﻿using System;
using System.Text;

namespace MyJScript
{
    /// <summary>
    /// Class for string object in MyJScript.
    /// </summary>
    public class MJSString : MJSObject
    {
        /// <summary>
        /// String value.
        /// </summary>
        string value;

        #region String primitives
        /// <summary>
        /// Primitive for blink
        /// </summary>
        /// <param name="this">this object</param>
        /// <param name="args">arguments list</param>
        /// <returns>return value</returns>
        private MJSObject PrimitiveBlink(MJSObject @this, params MJSObject[] args)
        {
            // Only on string
            MJSString s = @this as MJSString;
            if (s == null)
                return null;

            // Append HTML tag
            StringBuilder res = new StringBuilder("<blink>");
            res.Append(s.ValueOf() as string);
            res.Append("</blink>");

            return new MJSString(res.ToString());
        }

        /// <summary>
        /// Primitive for bold
        /// </summary>
        /// <param name="this">this object</param>
        /// <param name="args">arguments list</param>
        /// <returns>return value</returns>
        private MJSObject PrimitiveBold(MJSObject @this, params MJSObject[] args)
        {
            // Only on string
            MJSString s = @this as MJSString;
            if (s == null)
                return null;

            // Append HTML tag
            StringBuilder res = new StringBuilder("<bold>");
            res.Append(s.ValueOf() as string);
            res.Append("</bold>");

            return new MJSString(res.ToString());
        }

        /// <summary>
        /// Primitive for substr
        /// </summary>
        /// <param name="this">this object</param>
        /// <param name="args">arguments list</param>
        /// <returns>return value</returns>
        private MJSObject PrimitiveSubstr(MJSObject @this, params MJSObject[] args)
        {
            // Only on string
            MJSString s = @this as MJSString;
            if (s == null)
                return null;

            // Get index
            if (args.Length < 1)
                throw new NullReferenceException();
            int index = 0;
            if (args[0] is MJSNumber)
                index = (Int32)args[0].ValueOf();
            else if (args[0] is MJSString)
                index = Convert.ToInt32(args[0].ValueOf() as string);
            else
                throw new NotSupportedException();

            // Get size
            int size = (s.ValueOf() as string).Length-index;
            if (args.Length >= 2)
            {
                if (args[1] is MJSNumber)
                    size = (Int32)args[1].ValueOf();
                else if (args[1] is MJSString)
                    size = Convert.ToInt32(args[1].ValueOf() as string);
                else
                    throw new NotSupportedException();
            }

            // Get substr
            return new MJSString((s.ValueOf() as string).Substring(index, size));
        }
        #endregion

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="value">string value</param>
        public MJSString(string value)
        {
            this.value = value;
            base.Set("length", new MJSNumber(value == null ? 0 : value.Length));
            base.Set("blink", new MJSFunction(PrimitiveBlink));
            base.Set("bold", new MJSFunction(PrimitiveBold));
            base.Set("substr", new MJSFunction(PrimitiveSubstr));
        }

        /// <summary>
        /// Set a member value.
        /// </summary>
        /// <param name="name">member name</param>
        /// <param name="value">new value</param>
        public override void Set(string name, MJSObject value)
        {
            // Do nothing here
        }

        /// <summary>
        /// Get the value.
        /// </summary>
        /// <returns>value</returns>
        public override object ValueOf()
        {
            return value;
        }

        /// <summary>
        /// Convert to string.
        /// </summary>
        /// <returns>value as string</returns>
        public override string ToString()
        {
            if (value == null)
                return "null";

            StringBuilder res = new StringBuilder();
            res.Append('"');
            res.Append(value);
            res.Append('"');
            return res.ToString();
        }
    }
}
